(* $Id: godi_versions.ml 964 2012-05-25 23:41:21Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)


type condition =
    (* Contains only the version number, not the package name *)
    Exact of string    (* pkg-version or pkg==version *)
  | Any                (* pkg-* *)
  | Ge of string       (* pkg>=version *)
  | Gt of string       (* pkg>version *)
  | Le of string       (* pkg<=version *)
  | Lt of string       (* pkg<version *)
  | Ne of string       (* pkg!=version *)
  | Binary of string * string
;;

type dependency = string * condition   (* package name, version *)
;;    

let nv_re = Str.regexp "^\\(.*\\)-\\([0-9][^-]*\\)$";;

let godi_re = Str.regexp "godi\\([0-9]+\\)$";;

let split_vsuffix vsuffix =
  try
    let k = Str.search_forward godi_re vsuffix 0 in
    (String.sub vsuffix 0 k), int_of_string (Str.matched_group 1 vsuffix)
  with
      Not_found -> vsuffix, 0
  

let split_name_vsuffix fullname =
  if Str.string_match nv_re fullname 0 then (
    let n = Str.matched_group 1 fullname in
    let vr = Str.matched_group 2 fullname in
    (n, vr)
  )
  else 
    (* version is missing *)
    (fullname, "")
;;

let split_name_version fullname =
  let (n,vsuffix) = split_name_vsuffix fullname in
  let (v,rev) = split_vsuffix vsuffix in
  (n,v,rev)


let concat_name_vsuffix name vsuffix =
  if vsuffix = "" then
    name
  else
    name ^ "-" ^ vsuffix
  
let concat_vsuffix version pkgrevision =
  version ^ 
    (if pkgrevision > 0 then "godi" ^ string_of_int pkgrevision else "")


let concat_name_version name version pkgrevision =
  concat_name_vsuffix
    name
    (concat_vsuffix version pkgrevision)


let dep_any_re = Str.regexp "^\\(.*\\)-\\(\\[0-9\\]\\)?\\*$";;
  (* Either: name-* or name-[0-9]*. The latter is for backwards 
   * compatibility
   *)

let dep_op_re = Str.regexp "^\\(.*\\)\\(>=\\|>\\|<=\\|<\\|==\\|!=\\)\\([0-9][^-@]*\\)$";;

let dep_bin_re = Str.regexp "^\\(.*\\)-\\([0-9][^-@]*\\)@\\(.*\\)$" ;;

let bad_chars_re = Str.regexp ".*[*[?{]";;


let parse_dependency ?(sloppy=false) dep =
  (* Sloppy mode: also accept ambiguous notations to some extent (for
     command line)
   *)
  if Str.string_match dep_any_re dep 0 then (
    (Str.matched_group 1 dep, Any)
  )
  else if Str.string_match dep_op_re dep 0 then (
    let name =    Str.matched_group 1 dep in
    let op =      Str.matched_group 2 dep in
    let version = Str.matched_group 3 dep in
    (* Remove pkgrevision, if part of version: *)
    let (_,version,_) = split_name_version("x-" ^ version) in
    match op with
	">=" -> (name, Ge version)
      | ">"  -> (name, Gt version)
      | "<=" -> (name, Le version)
      | "<"  -> (name, Lt version)
      | "==" -> (name, Exact version)
      | "!=" -> (name, Ne version)
      | _    -> assert false
  )
  else if Str.string_match dep_bin_re dep 0 then (
    let name =    Str.matched_group 1 dep in
    let version = Str.matched_group 2 dep in
    let fprint  = Str.matched_group 3 dep in
    (name, Binary(version,fprint))
  )
  else (
    let (name, version, _) = split_name_version dep in
    (* TODO: Would like to raise User_error instead of Failure, but
     * Godi_types is a descendent module.
     *)
    if version = "" then (
      if sloppy then
	(name, Any)
      else
	failwith ("Cannot parse dependency: " ^ dep)
    ) else (
      if Str.string_match bad_chars_re version 0 then
	failwith ("Cannot parse dependency: " ^ dep);
      (name, Exact version)
    )
  )
;;


let print_dependency (name, cond) =
  match cond with
      Exact v -> concat_name_version name v 0
    | Any     -> name ^ "-*"
    | Ge v    -> name ^ ">=" ^ v
    | Gt v    -> name ^ ">"  ^ v
    | Le v    -> name ^ "<=" ^ v
    | Lt v    -> name ^ "<"  ^ v
    | Ne v    -> name ^ "!=" ^ v
    | Binary(v,fp) -> name ^ "-" ^ v ^ "@" ^ fp
;;


let any_dependency name =
  (name,Any)
;;


type version_element =
    (* A version_element suggests a certain weight of the corresponding
     * data
     *)
    [ `Sep of string
    | `Dec
    | `Nondec
    ]


let weight =
  function
    | `Sep "pl" -> 1
    | `Sep "_" -> 2
    | `Sep "+" -> 3
    | `Sep "." -> 4
    | `Sep "test" -> 10
    | `Sep "alpha" -> 11
    | `Sep "beta" -> 12
    | `Sep "pre" -> 13
    | `Sep "rc" -> 14
    | `Sep "" -> 15    (* neither of test ... rc *)
    | `Sep _ -> assert false
    | `Dec -> 20
    | `Nondec -> 30
;;


let sep_re = Str.regexp "\\.\\|\\+\\|_\\|pl";;
let dec_re = Str.regexp "\\([0-9]*\\)\\(test\\|alpha\\|beta\\|pre\\|rc\\|\\)";;


let parse_version_string v : (version_element * int) list =
  let pv = ref [] in
  let k = ref 0 in

  while !k < String.length v do
    if Str.string_match sep_re v !k then (
      let sep = Str.matched_string v in
      pv := (`Sep sep, 0) :: !pv;
      k := !k + String.length sep
    )
    else if Str.string_match dec_re v !k && Str.matched_string v <> "" then (
      let dec_str = Str.matched_group 1 v in
      let dec_val = if dec_str = "" then (-1) else int_of_string dec_str in
      let pre_str = Str.matched_group 2 v in
      pv := (`Sep pre_str, 0) :: (`Dec, dec_val) :: !pv;
      (* Attn: !pv is reversed! *)
      k := !k + String.length (Str.matched_string v);
    )
    else (
      pv := (`Nondec, Char.code(v.[ !k ])) :: !pv;
      incr k
    )
  done;

  List.rev !pv
;;


let cmp_version_string v1 v2 =
  let pv1 = ref (parse_version_string v1) in
  let pv2 = ref (parse_version_string v2) in
  
  let r = ref 0 in

  while !r = 0 && !pv1 <> [] && !pv2 <> [] do
    let (ve1, data1) = List.hd !pv1 in
    let (ve2, data2) = List.hd !pv2 in

    let w1 = weight ve1 in
    let w2 = weight ve2 in

    r := w1 - w2;

    if (!r == 0) then (
      r := data1 - data2;
    );

    pv1 := List.tl !pv1;
    pv2 := List.tl !pv2;
  done;

  match (!r, !pv1, !pv2) with
      0, [], [] -> 0
    | 0, [], _  -> (-1)
    | 0, _, []  -> 1
    | 0, _, _   -> assert false
    | n, _, _   -> n
;;


let version_matches_condition name cond =
  match cond with
    | Exact v -> name = v
    | Binary(v,_) -> name = v
    | Any     -> true
    | Ge v    -> cmp_version_string name v >= 0
    | Gt v    -> cmp_version_string name v >  0
    | Le v    -> cmp_version_string name v <= 0
    | Lt v    -> cmp_version_string name v <  0
    | Ne v    -> name <> v
;;


let create_binary_cond v fp = Binary(v,fp)
