(* $Id: menu_select.ml 1183 2013-04-14 13:02:56Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_query
open Printf


type case =
    [ `Source | `Binary ]

let string_of_case =
  function
    | `Source -> "source"
    | `Binary -> "binary"


let help_str =
  "This menu shows the list of software GODI knows something about. This \n" ^
  "includes installed software packages, and available software that is not\n" ^
  "yet compiled and installed. The INSTALLED and AVAILABLE columns are filled\n" ^
  "with the version numbers of installed and available packages, respectively.\n" ^
  "\n" ^
  "The FL column (flags) shows what is planned for every package: 'r' means\n" ^
  "that the installed software will be removed, 'b' means that the available\n" ^
  "software will be built. Often, both flags are set together, which means that\n" ^
  "a package is rebuilt (old version is removed, a new version is built).\n" ^
  "\n" ^
  "The flags can be changed in three ways:\n" ^
  "\n" ^
  "- By entering the number of the package (displayed in brackets), you can\n" ^
  "  invoke the package menu. This menu not only shows details of the software\n" ^
  "  but you can also select and deselect packages.\n" ^
  "\n" ^
  "- The upgrade function (letter 'u') selects all available packages that\n" ^
  "  are newer than the installed packages.\n" ^
  "\n" ^
  "- If you start the installation (letter 's'), GODI often selects further\n" ^
  "  packages automatically. This ensures that missing required software is\n" ^
  "  installed first, and that dependent software is rebuilt when necessary.\n" ^
  "  If GODI changes the selection, the selection list is shown again, and\n" ^
  "  the changes are displayed with emphasis. You can still edit the selection\n" ^
  "  list, and you can even undo your plan if you do not like the consequences\n" ^
  "  GODI suggests.\n" ^ 
  "\n" ^
  "By the way, your changes of the selection list are stored permanently in the\n" ^
  "file wishes.txt, so you can always exit the GODI console, and continue later.\n" ^
  "\n" ^
  "To actually install or remove packages, you must start the installation\n" ^
  "(letter 's'), and GODI will guide you through the build or removal process."
;;


let weight cat =
  match cat with
      `U -> 0
    | `D -> 1
    | `S -> 2
    | `I -> 3
;;


let abbr_full_version v rev =
  match rev with
      0 -> v
    | n -> v ^ "#" ^ string_of_int n
;;


let disp_status_str =
  function
      `Keep -> " "
    | `Build | `Build_dep _ -> "b"
    | `Remove | `Remove_dep _ -> "r"
    | `Add | `Add_dep _ -> "a"
;;


let disp_status sp_opt ip_opt =
(*
  let disp_status_ip() =
    match ip_opt with
	Some ip ->
	  disp_status_str (any_pkg ip)#user_status
      | None ->
	  disp_status_str `Keep
  in
 *)
  ( match ip_opt with
	Some ip -> disp_status_str (any_pkg ip)#user_status
      | None    -> " "
  ) ^
  ( match sp_opt with
	Some sp -> disp_status_str (any_pkg sp)#user_status
      | None    -> " "
  )
;;


let mk_sep_line width s =
  let l = String.length s in
  let p1 = 10 in
  String.make p1 '=' ^ s ^ String.make (min width (width-l-p1)) '='
;;


let lf_re = Str.regexp "\n";;


let forget_all_on_quit = ref false;;



let update_dependencies (pg : Miniui.page) iteration src_list inst_list =
  (* May raise User_error or Mission_impossible *)

  pg # clear();

  pg # configure ~auto_scroll:true ();

  Menu_aux.add_compact_head pg ">>> Dependency analysis <<<";

  let cleanup = ref (fun () -> ()) in

  (* Write a temporary godi.conf file: *)
  let tmp = Godi_file.allocate_tmpdir() in
  cleanup := (fun () -> Godi_file.release_tmpdir tmp);
  let tmp_path = Godi_file.tmpdir_path tmp in
  let tmp_godi_conf = Godi_file.addname tmp_path "godi.conf" in
  let mflag = Godi_conffile.modify_godi_conf 
		~pref_user_error:true ~destfile:tmp_godi_conf src_list in
  
  try
    let disp_proto = new Protocol.display_protocol ~auto_update:true pg in
    let protocol = disp_proto # collector in
    let ia = { Godi_exec.period = (-1.0);
	       Godi_exec.callback = (fun () -> true) } in
    let con = { protocol = protocol; interactivity = ia } in
    pg # set_busy true;
    
    let src_list' =
      if mflag then (
	(* There is a modified option, so we have to rescan: *)
	
	pg # put_line "Rescanning source packages for changed dependencies...";
	pg # update();
	
	forget_all_on_quit := true;
	let vid = new_var_id() in
	List.iter
	  (fun (pkg : package) -> 
	     (source_pkg pkg)#rescan_deps
	       con (Godi_file.slashname tmp_godi_conf) vid)
	  src_list;
	
        (* CHECK: What to do with errors? *)
	
	List.map 
	  (fun pkg -> `Source_pkg ((source_pkg pkg)#variant vid))
	  src_list
      )
      else
	src_list in
    
    pg # put_line "Analyzing dependencies...";
    
    let n =
      (* First try without upgrades (faster, better errors) *)
      try
        Godi_plan.update_plan 
          ~replace:`Never ~iteration (src_list' @ inst_list)
      with
        | Godi_plan.Mission_impossible msg1 ->
            try
              Godi_plan.update_plan 
                ~replace:`Upgrade ~iteration (src_list' @ inst_list)
            with
              | Godi_plan.Mission_impossible msg2 ->
                  let msg3 =
                    sprintf "Attempting without upgrades:\n%s\n \n\
                             Attempting with upgrades:\n%s"
                            msg1 msg2 in
                  raise (Godi_plan.Mission_impossible msg3)
    in
    (* May raise Mission_impossible *)
    
    (* Reset config_dialog_shown for the new `Build_dep packages: *)
    List.iter
      (function
	   `Source_pkg sp ->
	     (match sp # user_status with
		| `Build_dep i when i = iteration ->
		    sp # set_config_dialog_shown false
		| `Add_dep i when i = iteration ->
		    sp # set_config_dialog_shown false
		| _ ->
		    ()
	     )
	 | _ -> 
	     ()
      )
      src_list';
    
    !cleanup();
    (n, src_list')
  with
      error ->
	!cleanup();
	raise error
;;


type cat = [ `U | `D | `S | `I ];;
    (* `U: Unknown 
     * `D: Dependency selections
     * `S: Source/binary packages
     * `I: Installed packages that are not `S
     *)



let display_list (pg : Miniui.page) case avail_list_0 inst_list =

  let case_lc = string_of_case case in
  let case_uc = String.capitalize case_lc in

  let show_error errmsg =
    pg # clear();

    Menu_aux.add_compact_head ~lmargin:6 pg ">>> Error <<<";

    pg # put_line "The dependency analysis has resulted in an error.";
    pg # put_line "Error message: ";

    let disp_proto = new Protocol.display_protocol pg in

    disp_proto # add_string errmsg;
    
    forget_all_on_quit := true;

    ignore(pg # add_foot_button 
	     ~text:"ok"
	     ~underline:0
	     ~callback:(fun () -> pg # quit())
	     (`Keypress 'o'));

    pg # loop()
  in

  let upgrade combined_list () =
    let can_be_upgraded =
      List.filter
	(function 
	     (name,_,Some avail_pkg,Some inst_pkg) ->
	       newer_package avail_pkg inst_pkg
	   | _ ->
	       false
	)
	combined_list in

    let pg = pg # new_page() in
    pg # clear();

    Menu_aux.add_compact_head ~lmargin:6 pg ">>> Upgrade All <<<";

    if can_be_upgraded <> [] then (
      pg # put_line "For the following installed packages there are newer";
      pg # put_line "replacements:";
      pg # put_line "";
    )
    else
      pg # put_line "Nothing can be upgraded.";
    
    List.iter
      (function
	   (name,_,Some avail_pkg,Some inst_pkg) ->
	     let sp = (any_pkg avail_pkg) in
	     let ip = (any_pkg inst_pkg) in
	     let v_from = abbr_full_version ip#version ip#pkgrevision in
	     let v_to =   abbr_full_version sp#version sp#pkgrevision in
	     pg # put_line (sprintf "- %s-%s (installed: %s)" name v_to v_from);
	 | _ ->
	     ()
      )
      can_be_upgraded;

    ignore(pg # add_foot_button ~text:"ok" ~underline:0
             ~callback:(fun () ->
			  List.iter
			    (function
				 (name,_,Some avail_pkg,Some inst_pkg) ->
				   let sp = (any_pkg avail_pkg) in
				   let ip = (any_pkg inst_pkg) in
				   sp # set_user_status 
                                     (match case with
                                        | `Source -> `Build
                                        | `Binary -> `Add
                                     );
				   ip # set_user_status `Remove;
			       | _ ->
				   ()
			    )
			    can_be_upgraded;
			  pg # quit();
		       )
	     (`Keypress 'o'));

    ignore(pg # add_foot_button ~text:"exit menu" ~underline:1
             ~callback:pg#quit (`Keypress 'x'));

    pg # loop();
  in

  let cmp_name p1 p2 =
    let p = 
      compare (any_pkg p1)#name (any_pkg p2)#name in
    if p = 0 then  (* ensure stable sort *)
      compare p1 p2
    else
      p in

  let name_of p = (any_pkg p)#name in

  let cat_of p fallback : cat =
    match (any_pkg p)#user_status with
	`Build_dep _
      | `Remove_dep _ -> `D
      | _ -> fallback
  in

  let cat2_of p1 p2 : cat =
    match (any_pkg p1)#user_status with
      | `Build_dep _ -> `D
      | `Add_dep _   -> `D
      | `Build       -> `S
      | `Add         -> `S
      | _            -> cat_of p2 `S
  in


  let rec merge_packages (pl1 : package list) (pl2 : package list) =
    (* pl1, pl2 must already be sorted by name *)
    match (pl1, pl2) with
	(p1 :: pl1', p2 :: pl2') when name_of p1 = name_of p2 ->
	  (name_of p1, cat2_of p1 p2, Some p1, Some p2) :: 
	  merge_packages pl1' pl2'
      | (p1 :: pl1', p2 :: pl2') when name_of p1 < name_of p2 ->
	  (name_of p1, cat_of p1 `S, Some p1, None) :: 
	  merge_packages pl1' pl2
      | (p1 :: pl1', p2 :: pl2') (* when name_of p1 > name_of p2 *) ->
	  (name_of p2, cat_of p2 `I, None, Some p2) :: 
	  merge_packages pl1 pl2'
      | ( p1 :: pl1', [] ) ->
	  (name_of p1, cat_of p1 `S, Some p1, None) :: 
	  merge_packages pl1' []
      | ( [], p2 :: pl2' ) ->
	  (name_of p2, cat_of p2 `I, None, Some p2) :: 
	  merge_packages [] pl2'
      | ( [], [] ) ->
	  []
  in


  let package_list avail_list =
    (* Merge source and installed packages to triples:
     * (name, avail_opt, inst_opt)
     *)

    let combined_list = merge_packages 
			  (List.sort cmp_name avail_list)
			  (List.sort cmp_name inst_list) in

    (* Sort the packages:
     * - First `Build_dep and `Remove_dep packages, sorted by iteration number
     *   (however, `Remove_dep packages with a `Build source packages not)
     * - Then other source packages, sorted by name
     * - Then installed packages for which we have no source, by name
     *)
    let combined_list' =  (* = sorted combined_list *)
      List.sort
	(fun (n1,cat1,sp1_opt,ip1_opt) (n2,cat2,sp2_opt,ip2_opt) ->
	   let w1 = weight cat1 in
	   let w2 = weight cat2 in
	   match w1-w2 with
	       0 ->
		 (* Source packages first: *)
		 ( match sp1_opt, sp2_opt with
		       Some _, None -> (-1)
		     | None, Some _ -> 1
		     | _, _         -> 
			 (* Sort by name: *)
			 compare n1 n2
		 )
	     | d -> d
	)
	combined_list in
    combined_list'
  in

  let rec dep_head_of_list l =
    match l with
      | (_, `D, _, _) as x :: l' -> x :: dep_head_of_list l'
      | _ :: l' -> []
      | [] -> []
  in

  let names l =
    List.map
      (fun (name,_,_,_) -> name)
      l in

  let rec make_menu ?(scroll = 0.0) iteration avail_list =
    (* Create and display the menu: *)

    (* Note: make_menu has an src_list argument because the source objects
     * may be exchanged by different ones during the dependency analysis.
     *)

    let disp_list = package_list avail_list in
    let sel_array = Array.of_list disp_list in
    let dep_head = dep_head_of_list disp_list in
    
    pg # clear();

    Menu_aux.add_compact_head ~lmargin:6 pg 
      (">>> Select " ^ case_uc ^ " Packages <<<");

    ignore(pg # add_foot_button ~text:"upgrade all" ~underline:0
	     ~callback:(fun () ->
			  upgrade disp_list ();
			  make_menu (iteration+1) avail_list
		       ) 
	     (`Keypress 'u'));
    
    ignore(pg # add_foot_button 
	     ~text:"start/continue installation"
	     ~underline:0
	     ~callback:(fun () ->
			  (* Update the build plan: *)
			  try
			    let (n, avail_list') =
			      update_dependencies 
				pg iteration avail_list inst_list in
                            let new_disp_list = package_list avail_list' in
                            let new_dep_head = dep_head_of_list new_disp_list in
			    if names dep_head = names new_dep_head then (
			      (* Start installation: *)
			      let pg' = pg # new_page() in
			      Menu_install.display pg' avail_list' inst_list;
			      pg # quit();
			      (* Quit, so the packages are scanned again *)
			    )
			    else
			      make_menu (iteration+1) avail_list'
			  with
			      Godi_plan.Mission_impossible errmsg ->
				show_error errmsg;
				make_menu (iteration+1) avail_list
			    | User_error errmsg ->
				Menu_aux.error_page pg errmsg;
				make_menu (iteration+1) avail_list
		       )
	     (`Keypress 's'));

    ignore(pg # add_foot_button ~text:"help" ~underline:0
             ~callback:(Menu_help.display pg#new_page help_str)
             (`Keypress 'h'));
 
    ignore(pg # add_foot_button ~text:"exit menu" ~underline:1
             ~callback:pg#quit (`Keypress 'x'));

    let cat = ref (`U : cat) in
    
    let rev_menu = ref [] in

    for k = 0 to Array.length sel_array - 1 do
      let (name, new_cat, avail_pkg_opt, inst_pkg_opt) = sel_array.(k) in

      let avail_version_str =
	match avail_pkg_opt with
	    None -> ""
	  | Some p -> 
	      abbr_full_version (pkg_version p) (pkg_pkgrevision p)
      in

      let inst_version_str =
	match inst_pkg_opt with
	    Some ipkg -> 
	      let ip = any_pkg ipkg in
	      abbr_full_version ip#version ip#pkgrevision
	  | None   -> 
	      ""
      in
      
      (* Maybe pkg is member of a new category. If so, display a separating
       * line.
       *)
      if new_cat <> !cat then (
	let msg = 
	  match new_cat with
	      `U -> ""
	    | `D -> "Packages that have been (de)selected because of deps:"
	    | `S -> ("Packages available as " ^ case_lc ^ " code:")
	    | `I -> "Installed packages not available for reinstall:" in
	rev_menu := (`Separator (mk_sep_line (pg#width()) msg)) :: !rev_menu;
      );
      cat := new_cat;
      
      let eff_pkg =
	match avail_pkg_opt with
	    Some avail_pkg -> avail_pkg
	  | None ->
	      ( match inst_pkg_opt with
		    Some inst_pkg -> inst_pkg
		  | None -> assert false
	      )
      in

      let comment = (any_pkg eff_pkg)#comment in

      (* Add pkg to menu: *)
      rev_menu := `Item ( [ sprintf "[%3d]" (k+1);
			    disp_status avail_pkg_opt inst_pkg_opt;
			    name;
			    inst_version_str;
			    avail_version_str;
			    comment
			  ],
			  (k+1),
			  (fun () ->
			     let p = pg # scroll_position in
			     Menu_package.display 
			       ?inst_pkg:inst_pkg_opt pg#new_page 
                               avail_list eff_pkg case;
			     save_and_make_menu ~scroll:p iteration avail_list
			  ) 
			) :: !rev_menu
    done;
    
    Menu_aux.add_menu
      ~lmargin:0
      pg
      [ `Left 5; `Delim " "; `Left 2; `Delim " "; `Left 18; `Delim " "; `Left 10; `Delim " "; `Left 10 ; `Delim " "]
      ( `Head [ ""; "FL"; "NAME"; "INSTALLED"; "AVAILABLE"; "COMMENT" ] :: 
	  List.rev !rev_menu );

    pg # set_scroll_position scroll

  and save_and_make_menu ?scroll iteration avail_list =
    (* CHECK:
     * Once we have support for binary packages, we have to be more careful
     * here not to overwrite wishes of the other package type.
     *)
    let wl = Godi_wishes.wishes_of_packages(avail_list @ inst_list) in
    Godi_wishes.write_wishes ~pref_user_error:true wl;
    make_menu ?scroll iteration avail_list

  in

  try
    make_menu 0 avail_list_0;
    pg # loop()
  with
      User_error msg ->
	Menu_aux.error_page pg msg
;;



let display (pg : Miniui.page) case =
  pg # clear();

  pg # configure ~auto_scroll:true ();

  let case_lc = string_of_case case in
  let case_uc = String.capitalize case_lc in

  Menu_aux.add_full_head pg 
    (">>> Select " ^ case_uc ^ " Packages <<<");

  let disp_proto = new Protocol.display_protocol ~auto_update:true pg in
  let protocol = disp_proto # collector in
  let ia = { Godi_exec.period = (-1.0);
	     Godi_exec.callback = (fun () -> true) } in
  let con = { protocol = protocol; interactivity = ia } in

  forget_all_on_quit := false;
  let followup_action = ref (fun () -> ()) in

  ( try
      pg # put_line ~lmargin:10 
        ( match case with
            | `Source ->
                "Scanning system for available source packages..."
            | `Binary ->
                "Scanning system/servers for available binary packages..."
        );
      pg # set_busy true;
      pg # update();
      let available_list = 
        match case with
          | `Source ->
              scan_source_packages con 
          | `Binary ->
              scan_all_binary_packages con in

      (* Set the config_dialog_shown flag for all source packages: *)
      List.iter
	(function
	   | `Source_pkg sp ->
	       sp # set_config_dialog_shown true
	   | _ ->
	       ()
        )
	available_list;
      
      pg # put_line ~lmargin:10 "Scanning system for installed packages...";
      let inst_list = 
	get_installed_packages ~verbose:false con in
      
      (* Apply package wishes: *)
      pg # put_line ~lmargin:10 "Reading wish list...";
      pg # update();
      let wl = Godi_wishes.read_wishes ~pref_user_error:true () in
      Godi_wishes.apply_wishes_to_packages available_list wl;
      Godi_wishes.apply_wishes_to_packages inst_list wl;

      pg # set_busy false;
      followup_action := 
        (fun () -> display_list pg case available_list inst_list);

    with
	User_error msg ->
	  Godi_exec.collect protocol ("Error: " ^ msg ^ "\n");
	  pg # set_busy false;
  );

  let continue = ref true in

  if Buffer.length disp_proto#buffer > 0 then (
    pg # put_line "### There are error messages!";
    pg # put_line "### Packages might have been incorrectly scanned.";
    ignore(pg # add_foot_button
	     ~text:"save protocol"
	     ~underline:0
	     ~callback:(fun () ->
			  disp_proto # save_dialog())
	     (`Keypress 's'));
    ignore(pg # add_foot_button
	     ~text:"continue nevertheless"
	     ~underline:0
	     ~callback:pg#quit
	     (`Keypress 'c'));
    ignore(pg # add_foot_button 
	     ~text:"exit menu"
	     ~underline:1
	     ~callback:(fun () -> pg#quit(); continue:=false)
	     (`Keypress 'x'));
    pg # loop();
  );

  if not !continue then
    forget_all_on_quit := true;

  if !continue then
    !followup_action();

  if !forget_all_on_quit then (
    forget_source_packages();
    let con = Menu_aux.trivial_console pg in
    notify_about_change_of_installed_packages con `Reconcile;
  )
;;

