(* $Id: wtool_onet4.ml 1190 2015-01-13 00:44:35Z gerd $ *)

open Wtool_types
open Printf

let string_of_exn e =
  match e with
    | Failure msg -> 
	msg
    | Unix.Unix_error(code,fn,path) ->
	if path = "" then
	  sprintf "%s (function: %s)" 
	    (Unix.error_message code)
	    fn
	else
	  sprintf "%s: %s (function: %s)" 
	    path
	    (Unix.error_message code)
	    fn
    | e ->
	"Exception " ^ Netexn.to_string e


let config_pipeline p =
(* Nethttp_client.Debug.enable := true; *)
  let ftp_syn = Hashtbl.find Neturl.common_url_syntax "ftp" in
  let opts = p # get_options in
  let opts' =
    { opts with
        Nethttp_client.schemes = opts.Nethttp_client.schemes @
        [ "ftp", ftp_syn, Some 21, Nethttp_client.proxy_only_trans_id ]
    } in
  p # set_options opts';
  p # set_transport_proxy_from_environment
    ~insecure:true
    [ "http_proxy",  Nethttp_client.http_trans_id;
      "https_proxy", Nethttp_client.https_trans_id;
      "ftp_proxy",   Nethttp_client.proxy_only_trans_id
    ]


let use_web_proxy url =
  (* determines whether to use a web proxy for [url] *)
  try
    let p = new Nethttp_client.pipeline in
    config_pipeline p;
    p # proxy_type url <> None
  with
    | _ -> false


let http_handler ~timeout ~user ~proto base_url : wtool_handler =
  let fs =
    Nethttp_fs.http_fs
      ~config_pipeline:(
	fun p ->
	  config_pipeline p;
	  let opts = p # get_options in
	  let opts' =
	    { opts with
		Nethttp_client.connection_timeout = timeout;
	    } in
	  p # set_options opts';
      )
      ~streaming:true
      base_url in

  let pipeline = fs#pipeline in

  ( object(self)
      method base_url = base_url

      method list path =
	try
	  let files = fs # readdir [] path in
	  `Ok
	    (List.filter
	       (fun n -> n <> "" && n.[0] <> '.')
	       files
	    )
	with
	  | Nethttp_client.Http_protocol (Nethttp_client.Timeout _) ->
	      `Timeout
	  | Nethttp_client.Http_protocol e ->
	      `Error (string_of_exn e)
	  | e ->
	      `Error (string_of_exn e)

      method get tdir paths =
	(* don't use fs, but the pipeline directly *)
	let n = ref 0 in
	let rlist =
	  List.map
	    (fun path ->
	       let r = ref (`Error "unserved") in
	       let url = fs#translate path in
	       let call = new Nethttp_client.get url in
	       let local_filename =
		 Godi_file.addname
		   (Godi_file.tmpdir_path tdir)
		   ("wtool-" ^ string_of_int !n) in
	       let local_filename_n =
		 Godi_file.nativename local_filename in
	       incr n;
	       call # set_response_body_storage
		 (`File(fun () -> local_filename_n));
	       pipeline # add_with_callback
		 call
		 (fun rcall ->
		    match rcall#status with
		      | `Unserved -> assert false
		      | `Http_protocol_error (Nethttp_client.Timeout _) ->
			  r := `Timeout
		      | `Http_protocol_error e ->
			  r := `Error(string_of_exn e)
		      | `Redirection | `Client_error | `Server_error ->
			  r := `Error("HTTP code " ^ 
					string_of_int
					(rcall#response_status_code))
		      | `Successful ->
			  r := `Ok local_filename;
		 );
	       r
	    )
	    paths in
	Unixqueue.run pipeline#event_system;
	List.map (fun rref -> !rref) rlist

      method put local_filename path =
	`Error "No support for HTTP PUT in this version of godi_console - \
                hint: enable GODI_WTOOL_POWER"

      method api args =
        `Error "No support for API calls in this version of godi_console - \
                hint: enable GODI_WTOOL_POWER"
    end
  )

let ftp_handler ~timeout ~user base_neturl : wtool_handler =
  let base_neturl1 =
    Neturl.make_url
      ~scheme:"ftp"
      ?user
      ~host:(Neturl.url_host base_neturl)
      ~port:(try Neturl.url_port base_neturl with Not_found -> 21)
      ~path:(try Neturl.url_path base_neturl with Not_found -> [""] )
      (Neturl.url_syntax_of_url base_neturl) in
  let base_url =
    Neturl.string_of_url base_neturl1 in

  let with_tmpdir_fs tmpdir f =
    let fs =
      Netftp_fs.ftp_fs
	~config_client:(fun c ->
			  c # configure_timeout timeout
		       )
	~tmp_directory:(Godi_file.nativename(Godi_file.tmpdir_path tmpdir))
	~tmp_prefix:"wtooltmp-"
	~keep_open:true
	base_url in
    try
      let r = f fs in
      fs # close();
      r
    with
      | error ->
	  fs # close();
	  raise error in

  let with_fs f =
    let tmpdir = Godi_file.allocate_tmpdir() in
    try
      let r = with_tmpdir_fs tmpdir f in
      Godi_file.release_tmpdir tmpdir;
      r
    with
      | error ->
	  Godi_file.release_tmpdir tmpdir;
	  raise error in

  ( object(self)
      method base_url = base_url

      method list path =
	try
	  with_fs 
	    (fun fs ->
	       let files = fs # readdir [] path in
	       `Ok (List.filter
		      (fun n -> n <> "" && n.[0] <> '.')
		      files
		   )
	    )
	with
	  | Netftp_client.FTP_timeout _ ->
	      `Timeout
	  | error ->
	      `Error(string_of_exn error)

      method get tmpdir paths =
	try
	  with_tmpdir_fs
	    tmpdir
	    (fun fs ->
	       List.map
		 (fun path ->
		    try
		      let local =
			fs # read_file [ `Binary ] path in
		      `Ok (Godi_file.filename (local#filename))
			(* We don't call [close] ! *)
		    with
		      | Netftp_client.FTP_timeout _ ->
			  `Timeout
		      | error ->
			  `Error(string_of_exn error)
		 )
		 paths
	    )
	with
	  | Netftp_client.FTP_timeout _ ->
	      List.map (fun _ -> `Timeout) paths
	  | error ->
	      let code =
		`Error(string_of_exn error) in
	      List.map (fun _ -> code) paths


      method put local_filename path =
	try
	  with_fs 
	    (fun fs ->
	       let local =
		 ( object
		     method filename = Godi_file.nativename local_filename
		     method close() = ()
		   end
		 ) in
	       fs # write_file [ `Create; `Truncate; `Binary ] path local;
	       `Ok ()
	    )
	with
	  | Netftp_client.FTP_timeout _ ->
	      `Timeout
	  | error ->
	      `Error(string_of_exn error)
		
      method api args =
        `Error "No support for API calls in this version of godi_console - \
                hint: enable GODI_WTOOL_POWER"
    end
  )


let file_handler base_path : wtool_handler =
  let fs_lz = (* may result in an exception *)
    lazy(
      Netfs.local_fs
	~root:base_path
	()
    ) in
  ( object(self)
      method base_url = 
	Neturl.string_of_url
	  (Neturl.file_url_of_local_path base_path)

      method list path =
	try
	  let fs = Lazy.force fs_lz in
	  let files = fs # readdir [] path in
	  `Ok(List.filter
		(fun n -> n <> "" && n.[0] <> '.')
		files
	     )
	with
	  | e ->
	      `Error(string_of_exn e)

      method get tmpdir paths =
	let n = ref 0 in
	List.map
	  (fun path ->
	     try
	       let fs = Lazy.force fs_lz in
	       let local_filename =
		 Godi_file.addname
		   (Godi_file.tmpdir_path tmpdir)
		   ("wtool-" ^ string_of_int !n) in
	       Netchannels.with_out_obj_channel
		 (new Netchannels.output_channel
		    (Godi_file.open_out `Binary local_filename))
		 (fun out ->
		    Netchannels.with_in_obj_channel
		      (fs # read [ `Binary ] path)
		      (fun obj -> out # output_channel obj)
		 );
	       `Ok local_filename
	     with
	       | error ->
		   `Error(string_of_exn error)
	  )
	  paths

      method put local_filename path =
	try
	  let fs = Lazy.force fs_lz in
	  Netchannels.with_in_obj_channel
	    (new Netchannels.input_channel
	       (Godi_file.open_in `Binary local_filename))
	    (fun inp ->
	       Netchannels.with_out_obj_channel
		 (fs # write [ `Create; `Truncate; `Binary ] path)
		 (fun obj ->
		    obj # output_channel inp
		 )
	    );
	  `Ok ()
	with
	  | error ->
	      `Error(string_of_exn error)

      method api args =
        `Error "No support for API calls in this version of godi_console - \
                hint: enable GODI_WTOOL_POWER"
    end
  )
  

let wtool_factory() =
  let timeout = ref 30.0 in
  let cuser = ref None in
  ( object(self)
      method configure_timeout t = timeout := t
      method configure_user u = cuser := Some u
      method configure_store_passwords _ = ()
      method configure_localbase _ = ()
      method open_location base_url =
	if base_url <> "" && base_url.[0] = '/' then
	  (* Just interpret as local filename *)
	  file_handler base_url
	else
	  try
	    let u = 
	      Neturl.parse_url ~accept_8bits:true 
		(Neturl.fixup_url_string base_url) in
	    let have_user_in_url =
	      Neturl.url_provides ~user:true u in
	    let u0 =
	      Neturl.remove_from_url
		~user:true ~user_param:true ~password:true
		u in
	    let bu =
	      Neturl.string_of_url u0 in  (* to get consistent behavior *)
	    let proto = 
	      Neturl.url_scheme u in
	    let user =
	      if have_user_in_url then (
		if Neturl.url_user ~encoded:true u = "GODI_USER" then
		  !cuser
		else
		  Some(Neturl.url_user u)
	      )
	      else None in
	    match proto with
	      | "http" ->
		  http_handler ~timeout:!timeout ~user ~proto bu
	      | "ftp" ->
		  if user = None && use_web_proxy bu then
		    http_handler ~timeout:!timeout ~user ~proto:"ftp" bu
		  else
		    ftp_handler ~timeout:!timeout ~user u0
	      | "file" ->
		  let path = Neturl.local_path_of_file_url u0 in
		  file_handler path
	      | _ ->
		  failwith ("Unsupported URL scheme: " ^ base_url)
	  with
	    | Neturl.Malformed_URL ->
		failwith ("Bad URL: " ^ base_url)

      method url_help =
	[ "http://host/path     (no WebDAV support)";
	  "https://host/path    (no WebDAV support)";
	  "ftp://host/path";
	  "file:///path"
	]
	

      method have_auth = false

    end
  )
